__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
__version__ = "0.0.0"


def table(data, ipy=True, fmt=str, md=False):
  """Data tabulator

  Parameters
  ----------
  data  : 2D iterable
  ipy  : attempt to display HTML object
  fmt  : format function for each cell
  md  : return markdown instead of html
  """
  raw = "<table>\n <tr>\n  <td>" + "</td>\n </tr><tr>\n  <td>".join(
      "</td><td>".join(map(fmt, i)) for i in data) + "</td>\n </tr>\n</table>"
  if ipy:
    try:
      from IPython.display import HTML, display
    except ImportError:
      pass
    else:
      display(HTML(raw))
  if md:
    raw = "\n".join("|".join(map(fmt, i)) for i in data)
  return raw



import logging
from os import path
import warnings
with warnings.catch_warnings():
  #warnings.simplefilter(action='ignore', category=FutureWarning)
  warnings.filterwarnings('ignore', message="numpy.dtype size changed",
      category=RuntimeWarning, module="h5py")
  warnings.filterwarnings('ignore', message=".*`float` to `np.floating`",
      category=FutureWarning, module="h5py")
  import h5py

__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
__version__ = "0.1.1"


class H5Array(object):
  """Wrap (d : h5py.Dataset, f : h5py.File) so that
  wrapped.x == numpy.array(f[d['x']])
  @return[out] wrapped  :
  """
  from numpy import array

  def __init__(self, data, meta=None):
    self.data = data
    self.meta = meta

  def __getattr__(self, item):
    return getattr(self.data, item)

  def __len__(self):
    return len(self.data)

  def __getitem__(self, item):
    res = self.data.__getitem__(item)
    return res if self.meta is None else self.array(self.meta[res])

  def __str__(self):
    return str(self.data)


class H5Reader(object):
  def __init__(self, fname, mode='r', prefix=None, **kwargs):
    self.f = h5py.File(fname, mode=mode, **kwargs)
    bname = path.splitext(path.split(fname)[-1])[0]
    self.prefix = prefix if prefix is not None \
        else bname + '/' if bname in self.f.keys() else ''

  def __getattr__(self, item):
    log = logging.getLogger(__name__)
    try:
      key = self.prefix + item
      log.debug(key)
      res = self.f[key]
    except KeyError:
      key = self.prefix + '/' + item
      log.debug(key)
      res = self.f[key]

    if res.dtype == 'O':
      return H5Array(res, self.f)
    return H5Array(res)

  def keys(self):
    return self.f[self.prefix].keys() if self.prefix else self.f.keys()

  def __str__(self):
    try:
      return str(dict((k, str(getattr(self, k))) for k in self.keys()))
    except AttributeError:
      return str(self.f) + ':' + ','.join(self.keys())



try:
  import cPickle as pickle
except ImportError:
  import pickle

import logging

__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
__version__ = "0.3.0"


def memoise(ignore_kwargs=None, ignore_args=None):
  if ignore_kwargs is None:
    ignore_kwargs = []
  if ignore_args is None:
    ignore_args = []
  db = {}
  log = logging.getLogger(__name__)

  def wrapper(fn):
    def inner(*a, **kw):
      vs = tuple(str(v) for v in a if v not in ignore_args)
      ks = sorted(
          (str(k), str(v)) for k, v in kw.items() if k not in ignore_kwargs)
      key = vs + tuple(ks)
      res = db.get(key, None)
      if res is None:
        log.info("miss:" + fn.__name__)
        log.info("key:" + str(key))
        db[key] = res = fn(*a, **kw)
      else:
        log.info("hit:" + fn.__name__)
      return res
    return inner
  return wrapper


def load_db(shmDB):
  from io import open
  fname = "/dev/shm/" + shmDB
  db = {}
  try:
    db = pickle.load(open(fname, "rb"))
  except IOError as e:
    if "no such file" not in str(e).lower():
      log = logging.getLogger(__name__)
      log.warn("Could not load db. Possibly first time?")
      log.warn(str(e))
  return db


def save_db(shmDB, db):
  from io import open
  fname = "/dev/shm/" + shmDB
  pickle.dump(db, open(fname, "wb"), -1)


def hashForce(i):
  """returns hash(str(i)) if hash(i) fails"""
  try:
    return hash(i)
  except TypeError:
    return hash(str(i))


def memo_db(shmDB, ignore_kwargs=None, ignore_args=None):
  """memoise to /dev/shm/"""
  if ignore_kwargs is None:
    ignore_kwargs = []
  if ignore_args is None:
    ignore_args = []
  db = load_db(shmDB)
  log = logging.getLogger(__name__)

  def wrapper(fn):
    def inner(*a, **kw):
      vs = tuple(str(v) for v in a if v not in ignore_args)
      ks = sorted(
          (str(k), str(v)) for k, v in kw.items() if k not in ignore_kwargs)
      key = vs + tuple(ks)
      res = db.get(key, None)
      if res is None:
        log.info("miss:" + fn.__name__)
        log.info("key:" + str(key))
        db[key] = res = fn(*a, **kw)
        save_db(shmDB, db)
      else:
        log.info("hit:" + fn.__name__)
      return res
    return inner
  return wrapper



__all__ = ["cprof", "lprof"]
__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
__version__ = "0.0.1"


def cprof(f):
    from cProfile import Profile

    def inner(*args, **kwargs):
        profile = Profile()
        try:
            profile.enable()
            result = f(*args, **kwargs)
            profile.disable()
            return result
        finally:
            profile.print_stats(sort=1)
    return inner


def lprof(follow=None):
    from line_profiler import LineProfiler

    if follow is None:
        follow = []

    def inner(func):
        def profiled_func(*args, **kwargs):
            profiler = LineProfiler()
            try:
                profiler.add_function(func)
                for f in follow:
                    profiler.add_function(f)
                profiler.enable_by_count()
                return func(*args, **kwargs)
            finally:
                profiler.print_stats()
        return profiled_func
    return inner



#!/usr/bin/env python
"""nvsmi v%s
Usage:
  nvsmi [options] [<gpuNums>]

Arguments:
  <gpuNums>      : gpu numbers sans delimiters, eg 01234567
                   (default will select all).

Options:
  -v, --version  : print version number and exit.
  -h, --help     : print this help mesasge and exit.
  --sleep=<s>    : [default: 5:float].
  --log=<lvl>    : CRITICAL|ERROR|WARN(ING)|[default: INFO]|DEBUG|NOTSET

Example:
  nvsmi --sleep 1 023
%s
"""
from __future__ import division
from argopt import argopt
from tqdm import tqdm
from tqdm._utils import _term_move_up
from ..utils import sh, TqdmStream
import re
import logging
from time import sleep, strftime
import sys
__version__ = "0.4.2"
__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"


# user nice system idle iowait irq softirq
RE_CPU = re.compile(
    r"^cpu([0-9]+)" + r"\s+([0-9]+)" * 7 + ".*?$",
    flags=re.M)
RE_TOP_CPU = re.compile(r"^%Cpu\(s\)\s*:.*?$", flags=re.M)
RE_TOP_MEM = re.compile(r"^KiB Mem\s*:.*?$", flags=re.M)
RE_NUM = re.compile("([0-9.]+)")
RE_NUM_PLUS = re.compile("([0-9.]+\+?)")
RE_NVSMI_NUM_GPUS = re.compile(r"^Attached GPUs\s*:\s*([0-9]+)$", flags=re.M)
RE_NVSMI_GPUS = re.compile("^GPU [0-9:A-Z.]+$", flags=re.M)
RE_NVSMI = re.compile(
    r"^\s*(Total|Used|Gpu|Memory|Process ID" +
    r"|Used GPU Memory|GPU \w+ Temp)\s*:\s*(.+)$",
    flags=re.M)


def bstr2int(x):
  """13 KiB => 13*1024"""
  return eval(x.replace(' ', '')
              .replace("iB", '')
              .replace('G', "*1024M")
              .replace('M', "*1024K")
              .replace('K', "*1024"))


def pid2user(pid):
  res = sh(*("ps --no-headers -u -p %d" % pid).split()).split()
  try:
    return res[0]
  except IndexError:
    return str(pid)


class GPU(object):
  def __init__(self, num, stats):
    self.num = num
    self.mem = [0, 0]
    self.tmp = [0, 0]
    self.pids = []

    for (k, v) in stats:
      if k == "Gpu":
        self.use = int(v.replace(" %", ''))
      elif k == "Used" and not self.mem[0]:
        self.mem[0] = bstr2int(v)
      elif k == "Total" and not self.mem[1]:
        self.mem[1] = bstr2int(v)
      elif k == "GPU Current Temp":
        self.tmp[0] = int(v.replace(" C", ''))
      elif k == "GPU Slowdown Temp":
        self.tmp[1] = int(v.replace(" C", ''))
      elif k == "Process ID":
        self.pids.append(int(v))

  def __repr__(self):
    return (self.num, self.mem, self.tmp, self.use, self.pids)


def gpuStats(gpuNums):
  """
  >>> gpuStats([0, 2, 3])
  [(0, (1024, 8000), (62, 94), 92, (1337, 1339)), ...]

  # ID, MEM, TEMP, USE, PIDS
  """
  # log = logging.getLogger(__name__)
  try:
    res = sh(*("nvidia-smi -q -d MEMORY,UTILIZATION,TEMPERATURE,PIDS".split()))
  except OSError:
    return []
  gpu_stats = RE_NVSMI_GPUS.split(res)[1:]  # string stats per GPU

  if not gpuNums:
    assert len(gpu_stats) == int(RE_NVSMI_NUM_GPUS.findall(res)[0])
    return len(gpu_stats)

  stats = [GPU(num, RE_NVSMI.findall(gpu_stats[num])) for num in gpuNums]
  # log.debug('\n'.join(str(s) for s in stats))
  return stats


def cpuStats(doEach=False, delay=1):
  log = logging.getLogger(__name__)
  stats = None
  if doEach:
    res = sh("cat", "/proc/stat")
    # log.debug(res)
    stats = RE_CPU.findall(res)
    stats = [map(int, s) for s in stats]
  else:
    res = sh("top", "-bn2", "-d%d" % delay)
    # log.debug(res)
    cStats = RE_TOP_CPU.findall(res)[-1]
    mStats = RE_TOP_MEM.findall(res)[-1]
    log.debug((cStats, mStats))
    cStats = map(float, RE_NUM.split(cStats)[1::2])
    mStats = map(lambda x: int(x.replace('+', '0')) * 1024,
                 RE_NUM_PLUS.split(mStats)[1::2])
    log.debug((cStats, mStats))
    cStats.insert(0, "cpu")
    mStats.insert(0, "mem")
    return [cStats, mStats]
  return stats
  # log.debug(stats)


def run(args):
  log = logging.getLogger(__name__)
  cBars = [tqdm(desc=str(s[0]),
                total=sum(s[1:]),
                leave=False, dynamic_ncols=True,
                unit_scale=True, unit_divisor=1024,
                bar_format="{l_bar}{bar}|" + (
                    "{n_fmt}/{total_fmt}B" if s[0] == "mem" else "{postfix}"))
           for s in cpuStats()]
  log.debug(len(cBars))
  log.debug(args.gpuNums)
  gpuNums = map(int, sorted(args.gpuNums or range(gpuStats([]))))
  log.debug(gpuNums)
  gBars = [tqdm(desc=stat + str(num), total=total,
                leave=False, dynamic_ncols=True,
                unit_scale=True, unit_divisor=1024,
                bar_format="{l_bar}{bar}|" + (
                    "{n_fmt}/{total_fmt}B{postfix}C" if stat == "mem" else
                    "{postfix}"))
           for num in gpuNums
           for (stat, total) in (("gpu", 100), ("mem", 9e9))]
  log.debug(len(gBars))
  try:
    while True:
      for (i, s) in enumerate(cpuStats(delay=1)):
        if s[0] == "mem":
          cBars[i].n = s[2]
          cBars[i].total = s[1]
        else:
          idl = s[4]
          tot = sum(s[1:])
          cBars[i].n = tot - idl
          cBars[i].total = tot
        if not i:
          cBars[i].set_postfix(time=strftime("%d/%H:%M:%S"))
        cBars[i].refresh()

      for gpu in gpuStats(gpuNums):
        i = gpuNums.index(gpu.num) * 2
        # tqdm.write(str([i, num, mem, tmp, use]))
        gBars[i + 1].n = gpu.mem[0]
        gBars[i + 1].total = gpu.mem[1]
        users = ', '.join(set([pid2user(p) for p in gpu.pids]))
        gBars[i + 1].set_postfix_str(str(gpu.tmp[0]), refresh=True)
        gBars[i].n = gpu.use
        gBars[i].set_postfix_str(users, refresh=True)
        gBars[i].refresh()

      sleep(args.sleep - 1)

  except KeyboardInterrupt:
    [b.close() for b in reversed(gBars)]
    [b.close() for b in reversed(cBars)]
    # sys.stderr.write('\n' * (len(gpuNums) * 2 - 1))
    sys.stderr.write(_term_move_up())


def main(argv=None):
  """argv  : list, optional (default: sys.argv[1:])"""
  args = argopt(__doc__ % (__version__, __author__),
                version=__version__).parse_args(args=argv)
  logging.basicConfig(level=getattr(logging, args.log, logging.INFO),
                      stream=TqdmStream)
  log = logging.getLogger(__name__)
  log.debug(args)
  return run(args) or 0


if __name__ == "__main__":
  main()



import logging
__all__ = ["Plt", "annotate_plot"]
__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
__version__ = "0.7.0"


class Plt(object):
  """
  Manage switching between subplots and individual figures with ease
  """
  import matplotlib.pyplot as plt
  import re
  _dims = None
  _f = 0  # current figure number (number of calls to set_dims)
  _n = 1  # current (sub) figure number (number of calls to fig)
  disable = False
  RE_NPATH = re.compile("[^A-Za-z0-9_!&()-+., ]+")

  @classmethod
  def set_dims(cls, rows=None, cols=None, fig=None):
    if cls.disable:
      return
    cls._dims = [rows, cols] if rows and cols else None
    # select previous figure if going back to figure mode,
    # else create a new figure to hold subplots
    cls._f = fig or (cls._f + 1 if cls._dims else cls._f)
    cls.plt.figure(cls._f)
    cls._n = 1

  @classmethod
  def fig(cls, n=None, **kwargs):
    if cls.disable:
      return
    if cls._dims is None:
      cls._f = cls._f + 1 if n is None else max(cls._f, n)
      cls.plt.figure(cls._f)
      cls._n = 1
    else:
      if n is None:
        n = cls._n
        cls._n += 1
      d = cls._dims + [n]
      cls.plt.subplot(*d)
    for (k, v) in kwargs.items():
      getattr(cls.plt, k)(v)

  @classmethod
  def saveax(cls, ext=".png", size=None, dpi=600):
    log = logging.getLogger(__name__)
    ax = cls.plt.gca()
    if size is not None:
      cls.plt.gcf().set_size_inches(size)
    curTitle = ax.get_title()
    log.debug("plot title: " + curTitle)
    ax.set_title("")
    opath = '_'.join(cls.RE_NPATH.sub("", curTitle).strip().split()) + ext
    log.debug("saving to " + opath)
    cls.plt.savefig(opath, dpi=dpi)
    log.debug("saved")
    ax.set_title(curTitle)

  @classmethod
  def draw(cls):
    if cls.disable:
      return
    # log = logging.getLogger(__name__)
    # cls.plt.tight_layout()
    cls.plt.draw()
    cls.saveax()

  @classmethod
  def trim(cls, im, vmin=None):
    """@return (yMin, yMax, xMin, xMax)"""
    if vmin is None:
      vmin = im.min()
    xs = (im.max(0) > vmin)
    ys = (im.max(1) > vmin)
    xMin = max(0, xs.argmax() - 5)
    yMin = max(0, ys.argmax() - 5)
    xMax = min(len(xs), len(xs) + 5 - xs[::-1].argmax())
    yMax = min(len(ys), len(ys) + 5 - ys[::-1].argmax())
    return yMin, yMax, xMin, xMax

  @classmethod
  def imshow(cls, im, xlabel=None, ylabel=None, cmap="hot", trim=True):
    if cls.disable:
      return
    log = logging.getLogger(__name__)
    vmin, vmax = im.min(), im.max()
    vmean, vstd = im.mean(), im.std()
    imDbgStr = '/'.join(["{:3.2e}"] * 4).format(vmin, vmean, vmax, vstd)
    log.debug("image min/mean/max/std" + imDbgStr)
    if log.getEffectiveLevel() <= logging.DEBUG:
      ax = cls.plt.gca()
      curTitle = ax.get_title()
      ax.set_title(curTitle + ' ' + imDbgStr)
    yMin, xMin = 0, 0
    yMax, xMax = im.shape
    if trim:
      yMin, yMax, xMin, xMax = cls.trim(im, vmin=vmin)

    cls.plt.imshow(im[yMin:yMax, xMin:xMax],
                   aspect="equal", origin="lower", cmap=cmap,
                   vmin=vmin, vmax=vmax)
    if not (xlabel or ylabel):
      cls.plt.axis("off")
    else:
      # cls.plt.axis("on")
      ax = cls.plt.gca()
      ax.tick_params(axis="both", which="both",
                     labelbottom="off", labelleft="off",
                     bottom=0, left=0, top=0, right=0)
      ax.set_frame_on(False)
      if xlabel:
        cls.plt.xlabel(xlabel)
      if ylabel:
        cls.plt.ylabel(ylabel)
    cls.draw()

  @classmethod
  def __getattr__(cls, name):
    if cls.disable:
      return
    return getattr(cls.plt, name)


def annotate_plot(x, y, names, *plot_args, **plot_kwargs):
    """
    Interactively annotated plots.

    Calls `matplotlib.pyplot.plot(x, y, *plot_args, **plot_kwargs)`,
    and to annotate:
    if `names` is `callable`: `names(x[i], y[i], i)`
    else: `names[i]`
    """
    import matplotlib.pyplot as plt
    line, = plt.plot(x, y, *plot_args, **plot_kwargs)
    ax = plt.gca()
    fig = plt.gcf()
    annot = ax.annotate("", xy=(0,0), xytext=(-20,20),textcoords="offset points",
                    bbox=dict(boxstyle="round", fc="w"),
                    arrowprops=dict(arrowstyle="->"))
    annot.set_visible(False)

    def update_annot(ind):
        x, y = line.get_data()
        ind = ind["ind"]
        annot.xy = (x[ind[0]], y[ind[0]])
        if callable(names):
            text = " & ".join(names(x[i], y[i], i) for i in ind)
        else:
            text = " & ".join(names[i] for i in ind)
        annot.set_text(text)
        annot.get_bbox_patch().set_alpha(0.25)

    def hover(event):
        vis = annot.get_visible()
        if event.inaxes == ax:
            cont, ind = line.contains(event)
            if cont:
                update_annot(ind)
                annot.set_visible(True)
                fig.canvas.draw_idle()
            else:
                if vis:
                    annot.set_visible(False)
                    fig.canvas.draw_idle()

    fig.canvas.mpl_connect("motion_notify_event", hover)



#!/usr/bin/env python
"""file v%s

Prints info about given files.

Usage:
  file [options] <files>...

Options:
  --log LEVEL  : (FAT|CRITIC)AL|ERROR|WARN(ING)|[default: INFO]|DEBUG|NOTSET

Arguments:
  <files>  : Files to test.

%s
"""
import sys
from io import open as io_open
import logging
from argopt import argopt
import re
__all__ = ["main", "run"]
__version__ = "0.1.0"
__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"

RE_PARTS = re.compile(r"[[\]:]")
TEXT_EXT = "txt md rst py cpp cfg ini yml yaml xml csv tsv".split()

def run(args):
  """@param args: RunArgs"""
  log = logging.getLogger(__name__)

  info = dict((f, None) for f in args.files)
  for f in info:
    parts = filter(None, RE_PARTS.split(f))
    fn = parts[0]
    ext = fn.rsplit('.')[-1]
    if ext == fn:
      ext = ""
    inds = [i.decode('U8') for i in parts[1:]]
    inds = [eval(i) if i.isnumeric() else i for i in inds]
    if ext == "mat":
      from caspyr.utils import H5Reader
      log.debug("hdf5 file")
      data = H5Reader(fn, prefix=inds[0] if len(inds) else None)
      for i in inds[1:]:
        data = data[i] if isinstance(i, int) else getattr(data, i)
    elif ext in TEXT_EXT:
      with io_open(fn) as fd:
        data = fd.read()
      if inds:
        # line
        data = data.split('\n')[inds[0] - 1]
        if inds[1:]:
          # char / cell
          i = inds[1] - 1
          if ext == "csv":
            data = data.split(',')[i]
          elif ext == "tsv":
            data = data.split('\t')[i]
          else:
            data = data[i]
          if inds[2:]:
            log.warn("ignoring:%r" % inds[2:])
    else:
      log.warn("unknown extension:%s:%s" % (ext, fn))
      data = None
    info[f] = data
  if log.getEffectiveLevel() <= logging.INFO:
    print('\n'.join("{:>20} : {}".format(f[:20], d) for f, d in info.items()))
  return info


def main(argv=None):
  """argv  : list, optional (default: sys.argv[1:])"""
  args = argopt(__doc__ % (__version__, __author__),
                version=__version__).parse_args(args=argv)
  logging.basicConfig(level=getattr(logging, args.log, logging.INFO))
  log = logging.getLogger(__name__)
  log.debug(args)
  return run(args) or 0


if __name__ == "__main__":
  main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""pandoc_filter v%s
Replaces:
  `figure`   : `figure*` if width=\textwidth
  `\[ ... \] : equation

Usage:
  pandoc_filter [--help|options] <type> [<file>]

Options:
  -h, --help     : Print this help and exit
  -v, --version  : Print module version number and exit
  --no-deqn      : bool, don't subsitiute \[\] -> \begin{equation}\end{equation}
  --MMD=<fmt>    : Output Makefile include information for DEPENDANT to FILE,
                   where `<fmt>` is `FILE:DEPENDANT`
  -p=<P>, --image-prefix=<P>  : for `--MMD`, e.g. "images/" [default: "":str]
  --log=<lvl>    : CRITICAL|ERROR|WARN(ING)|[default: INFO]|DEBUG|NOTSET

Arguments:
  <type>  : inplace|fromfile|(default: stdin|-)
      default (if unknown) is `stdin`. `fromfile` writes to `stdout`
  <file>  : *|(default: -).md|html|(default: latex)

%s
"""
import sys
import re
from argopt import argopt
import logging
__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
__all__ = ["main"]
__version__ = "0.3.2"


RE_TEX_WFIG = re.compile(
    r"(\\begin\{figure)(\})(\s*\\centering\s*" +
    r"\\includegraphics\[width=1\.?[0]*\\textwidth\].*?\\end\{figure)(\})",
    flags=re.M | re.S)
RE_TEX_DEQN = re.compile(r"\\\[([\s\S]*?)\\\]", flags=re.M)
RE_MD_NOTES = re.compile(r"^[Nn]otes?:\s*(\S.*?)\n\n", flags=re.M | re.S)
RE_LONGTAB_COL = re.compile(
    r"(\\begin\{minipage\}\[)t(\]\{[0-9.]+\\columnwidth\})", flags=re.M)
# sed -rn 's/.*!\[.*\]\(([^):]+)\).*/\1/p' "$<" | xargs echo "$@": > "$(<:%.md=$(DOUT)/%.d)"
RE_MD_IMG = re.compile(r"\!\[.*?\]\(\s*([^#][^):]*?)\)", flags=re.M | re.S)
RE_HTML_IMG = re.compile(r'<img\b.*src="([^":]+?)"', flags=re.M | re.S)
# sed -rn 's/.*\\includegraphics[^{]*\{([^}]+)\}.*/images\/\1/p' "$<" | xargs echo "$@": > "$(<:%.latex=%.d)"
RE_TEX_IMG = re.compile(
    r"\\includegraphics(\[[^]]*\])?\{([^}]+)\}", flags=re.M | re.S)
RE_TEX_HTML_IMG = re.compile(
    r"(\\includegraphics(\[[^]]*\])?\{\s*)(https?://[^}]+)(\})",
    flags=re.M | re.S)
RE_MD_REFS = re.compile(r" +(\[@.*?\])", flags=re.M)
RE_HTML_TEXT_TAGS = re.compile(r"<(?P<id>small|tiny)>(.*?)</(?P=id)>",
    flags=re.M | re.S)

# de-unicode
REPL = [i.split() for i in '''
“ "
” "
' '
' '
– --
'''.strip('\n').split('\n')]

def run(args):
  log = logging.getLogger(__name__)

  if not args.file:
    args.file = '-.latex'
  basename, ext = [s.lower() for s in args.file.rsplit('.', 1)]
  if not ext:
    ext = "latex"
  log.info("ext:" + ext)

  # READ
  fp = open(args.file) if args.type in ["inplace", "fromfile"] \
      else sys.stdin
  raw = fp.read().replace("\r\n", "\n")
  fp.close()

  # FILTER
  if ext == "md":
    log.info("slide notes")
    raw = RE_MD_NOTES.sub(r'<div class="notes">\n\n\1\n\n</div>\n\n', raw)
    log.info("nbsp refs")
    raw = RE_MD_REFS.sub(r"&nbsp;\1", raw)
    if args.MMD:
      log.info("MMD:" + args.MMD)
      fn, dependant = args.MMD.split(':')
      with open(fn, "w") as fd:
        images = RE_MD_IMG.findall(raw) + RE_HTML_IMG.findall(raw)
        log.debug(images)
        if args.image_prefix:
          images = [i if i.startswith(args.image_prefix)
                    else (args.image_prefix + i)
                    for i in images]
        fd.write(dependant + ": " + ' '.join(images))
  elif ext == "html":
    log.info("TeX-AMS_~~C~~HTML")
    raw = raw.replace("TeX-AMS_CHTML", "TeX-AMS_HTML")
  elif ext == "latex":
    log.info("wide figures")
    log.debug('\n---\n'.join(''.join(fig) for fig in RE_TEX_WFIG.findall(raw)))
    raw = RE_TEX_WFIG.sub(r"\1*\2\3*\4", raw)
    if not args.no_deqn:
      log.info("disp eqn's")
      raw = RE_TEX_DEQN.sub(r"\\begin{equation}\1\\end{equation}", raw)
    log.info("longtab")
    raw = RE_LONGTAB_COL.sub(r"\1\2", raw)
    # raw = RE_TEX_WFIG.sub(r"\\begin{strip}\1*\2[htbp]\3*\4\\end{strip}", raw)
    log.info("text tags")
    rawNew = RE_HTML_TEXT_TAGS.sub(r"{\\\1 \2}", raw)
    while raw != rawNew:
      raw = rawNew
      rawNew = RE_HTML_TEXT_TAGS.sub(r"{\\\1 \2}", raw)
    log.info("web images")
    for i in RE_TEX_HTML_IMG.findall(raw):
      log.warn("IGNORING:" + i[2])
    raw = RE_TEX_HTML_IMG.sub('', raw)
    log.info("ASCIIfy")
    for i, j in REPL:
      raw = raw.replace(i, j)
    if args.MMD:
      log.info("MMD:" + args.MMD)
      fn, dependant = args.MMD.split(':')
      with open(fn, "w") as fd:
        images = [im for (_, im) in RE_TEX_IMG.findall(raw)]
        log.debug(images)
        if args.image_prefix:
          images = [i if i.startswith(args.image_prefix)
                    else (args.image_prefix + i)
                    for i in images]
        fd.write(dependant + ": " + ' '.join(images))
  else:
    log.warn("ext not recognised")

  # WRITE
  if args.type == "fromfile" or basename in ('-', "stdout"):
    fp = sys.stdout
  else:
    fp = open(args.file, mode='w')
  fp.write(raw)
  fp.close()


def main(argv=None):
  """argv  : list, optional (default: sys.argv[1:])"""
  args = argopt(__doc__ % (__version__, __author__),
                version=__version__).parse_args(args=argv)
  logging.basicConfig(level=getattr(logging, args.log, logging.INFO))
  log = logging.getLogger(__name__)
  log.debug(args)
  return run(args) or 0


if __name__ == "__main__":
  main()



from __future__ import division
import logging
import matplotlib.pyplot as plt
import ipywidgets as ipyw
import math
__all__ = ["volshow"]
__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
__version__ = "0.0.0"

def volshow(vols, sharex=True, sharey=True, z=0, colorbar=True):
    """
    Jupyter notebook interactive volume display.

    @param vols  : [("title", 3darray), ...] or {"title": 3darray} or [3darray, ...].
      N.B.: `3darray` is `zyx`.
    @return: fig, axs
    """
    if isinstance(vols, dict):
        vols = vols.items()
    elif hasattr(vols, "shape") or hasattr(vols[0], "shape"):  # 4darray
        vols = [("", i) for i in vols]

    W = min(4, int(math.ceil(len(vols) ** 0.5)))
    H = int(math.ceil(len(vols) / W))
    assert W * H >= len(vols)
    fig = plt.figure()
    axs = [[None] * W for _ in range(H)]

    def plot_images(z=z):
        """z  : int, slice index"""
        plt.clf()

        axs[0][0] = plt.subplot(H, W, 1)
        plt.imshow(vols[0][1][z])
        if colorbar:
            plt.colorbar()

        for i, (title, vol) in enumerate(vols):
            if not i:
                continue
            axs[i // W][i % W] = plt.subplot(
                H, W, i + 1, sharex=axs[0][0] if sharex else None, sharey=axs[0][0] if sharey else None)
            plt.imshow(vol[z])
            plt.title(title)
            if colorbar:
                plt.colorbar()
        plt.show()

    ipyw.interact(plot_images,
        z=ipyw.IntSlider(min=0, max=len(vols[0][1]) - 1, step=1, value=len(vols[0][1])//2))
    return fig, axs



#!/usr/bin/env python
"""REgexMove v%s
Usage:
  remv [options] <ipattern> <opattern>

Options:
  -n, --dry_run  : only print commands
  -r, --recurse  : recurse into subdirectories
  --log=<lvl>    : CRITICAL|ERROR|WARN(ING)|[default: INFO]|DEBUG|NOTSET
%s
"""
from argopt import argopt
import logging
import re
from os import walk as os_walk
from os import path as os_path
from shutil import move as shutil_move
__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
__all__ = ["main"]
__version__ = "0.1.2"


def run(args):
  log = logging.getLogger(__name__)
  if args.recurse:
    allFiles = [os_path.join(d, f) for (d, _, fs) in os_walk('.')
                for f in fs]
  else:
    allFiles = os_walk('.').next()[2]
  RE_IPAT = re.compile(args.ipattern)
  for f in allFiles:
    if RE_IPAT.search(f):
      dst = RE_IPAT.sub(args.opattern, f)
      log.info(' '.join(("mv", f, dst)))
      if not args.dry_run:
        shutil_move(f, dst)


def main(argv=None):
  """argv  : list, optional (default: sys.argv[1:])"""
  args = argopt(__doc__ % (__version__, __author__),
                version=__version__).parse_args(args=argv)
  logging.basicConfig(level=getattr(logging, args.log, logging.INFO))
  log = logging.getLogger(__name__)
  log.debug(args)
  return run(args) or 0


if __name__ == "__main__":
  main()



from __future__ import print_function
from caspyr.multiprocessing import UnorderedQMap, Queue, Manager, \
    freeze_support
from time import sleep
from random import random


def stringify(*a, **k):
  return str((a, k))


def dummyFunc(resVec, i, resLen, **kwargs):
  assert i < resLen
  sleep(random() / 10)
  resVec[i] = "index %d says %s" % (i, str(kwargs))
  return i


def test_basic():
  """Test shortest canonical example"""
  inQ, outQ = Queue(), Queue()
  for i in range(10):
    inQ.put((stringify, (i, "say"), dict(hello="world"[i % 5])))

  with UnorderedQMap(outQ, inQ):
    for i in range(10):
      print(outQ.get())


def test_basic2():
  """Test managed list task despatch"""
  NUM_TASKS = 20

  # Create queues
  taskQ = Queue()
  resIdxQ = Queue()
  manager = Manager()
  results = manager.list([None] * NUM_TASKS)

  # Submit tasks
  [taskQ.put((dummyFunc, (results, i, NUM_TASKS), {"what": i % 2}))
   for i in range(NUM_TASKS)]

  # Start worker processes
  with UnorderedQMap(resIdxQ, taskQ):
    # Get and print results
    print("Unordered results:")
    for _ in range(NUM_TASKS):
      print(results[resIdxQ.get()])


if __name__ == "__main__":
  freeze_support()
  test_basic()
  test_basic2()



#!/usr/bin/env python
"""Plots 3d series from a csv file, including titles and colours.
Usage:
  plot3d [--help|options] <files>...

Options:
  -h, --help         Print this help and exit.
  -v, --version      Print module version number and exit.
  -a, --fix-aspect   Fix aspect ratio to be 1 [default: False].
  -d, --depthshade   Add alpha to far away points [default: False].
  -t, --timeseries   Interpret the last column of data as time,
                     plotting additional graphs.
  --colours=<wheel>  Colour wheel to use for differentiating between series
                     [default: rgbcmyk].
  --trans-mod=<t>    Custom transformation module containing `my_trans()`
                     (default: None). See `pitch_transforms` for syntax.
  -r, --right        Whether to set the origin and the right of the pitch.
                     Passed as an argument to <t>.
                     If [default: False], sets the origin at the left.
  --save-mod=<s>     Custom save module containing `my_save()`
                     (default: None). See `pitch_transforms` for syntax.
  --output-path=<p>  Directory in which to save generated figures
                     (default: None).
  --log=<lvl>        CRITICAL|ERROR|WARN(ING)|[default: INFO]|DEBUG|NOTSET.

Arguments:
  <files>  CSV files. Can include `-` (stdin).
           Line commands begin with  "#". The second character determines
           the command:

             #%Figure title (MANDATORY)
             #@Series title (MANDATORY)
             # Comment

           3D points are specified simply as:

             1.0,2,-3.141592653589793

           Symmetric x/y/z errors may optionally be specified:

             0.5,0.6,0.3,0.25,0.3,0.15

           Series titles can be postfixed with {.options}, as follows:

             {.plot_type=scatter}  (DEFAULT: scatter)
           | {.plot_type=plot}
             {.c=colour}           (DEFAULT: auto, from --colours=<wheel>)
             {.depthshade=}        (DEFAULT: auto, from --depthshade)
           | {.depthshade=1}
             {.s=<point_size>}     (DEFAULT: 32.0)
             {.alpha=<alpha>}      (DEFAULT: 1.0)
             {.xlabel=<x>}         (DEFAULT: x)

           Errorbar options:

             {.linestyle=<s>}      (DEFAULT: : if plot_type==plot else None)
             {.marker=<m>}         (DEFAULT: o if plot_type==scatter else None)

           Figure options:

             {.projection=3d}      (DEFAULT: 3d)
           | {.projection=2d}

           Any other options are forwarded to matplotlib.

           If y and/or z are unspecifed, they are assumed zero.
           A fourth integer coordinate may be specified for indexing.
           You can create multiple figures
           (TODO: resets colour wheel index):

             #%Another figure
"""
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # NOQA
from ..plotting import errorbar3d
from importlib import import_module
import sys
import re
import os.path
import logging
from docopt import docopt
__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
__all__ = ["main"]
__version__ = "0.18.2"


RE_options = re.compile(r'\s*\{\.(\w+)=(.*?[^\\])\}')


def get_options(s):
  """
  >>> get_options('foo {.bar=baz} {.bat=3}')
  {'bat': '3', 'bar': 'baz'}
  """
  return dict((k, v.replace(r"\}", "}").replace(r"\ ", ""))
              for (k, v) in RE_options.findall(s))


def strip_options(s):
  """
  >>> strip_options('foo {.bar=baz} {.bat=3}')
  'foo'
  """
  return RE_options.sub('', s)


def coord(m, i):
  """
  i^{th} column of a row-major matrix m. (Equivalently,
  i^{th} row of a column-major matrix m.)

  Parameters
  ----------
  m  : list of lists
  i  : int

  Returns
  -------
  out  : list
      i^{th} element of each list in m

  Example
  -------

  >>> coord([[0, 1, 2], [4, 1, 9], [6, 1, 5]], 1)
  [1, 1, 1]
  """
  return [v[i] if (0 <= i < len(v)) else 0 for v in m]


def save_csv(series, fname='plot3d_out.csv'):
  # TODO: errors
  log = logging.getLogger()
  log.info("Saving to " + fname)
  with open(fname, 'w') as fp:
    fig_names = []
    for series_name in sorted(series.keys()):
      if series_name[2] not in fig_names:
        fig_names.append(series_name[2])
        fp.write('#%' + series_name[2] + '\n')

      fp.write('\n#@' + series_name[1] + '\n')
      fp.write('\n'.join(','.join('{0:.3f}'.format(i).replace('.000', '')
                                  for i in xyz)
                         for xyz in series[series_name]) + '\n')


def fn_prepend(f, p='', s=''):
    r"""prefix/suffix basename within a full file path
    f  : full file path
    p  : prefix
    s  : suffix (before extension, if exists)

    Example
    -------

    >>> fn_prepend('C:\\path\\to.a\\file...with_an.ext', 'foo_', '_bar')
    'C:\\path\\to.a\\foo_file...with_an_bar.ext'

    """
    ftree = list(os.path.split(f))
    basename = '.'.join(ftree[-1].split('.')[:-1])
    ext = ftree[-1].split('.')[-1] if ftree[-1].count('.') > 0 else ''
    ftree[-1] = p + basename + s + '.' + ext
    return os.path.join(*ftree)


def axisEqual3D(ax3d):
  """Sets aspect ratio == (1,1) for a 3D axis"""
  import numpy as np
  extents = np.array([getattr(ax3d, 'get_' + dim + 'lim')() for dim in 'xyz'])
  sz = extents[:, 1] - extents[:, 0]
  centres = np.mean(extents, axis=1)
  r = max(abs(sz)) / 2
  for ctr, dim in zip(centres, 'xyz'):
    getattr(ax3d, 'set_' + dim + 'lim')(ctr - r, ctr + r)


def get(l, i, default_val=None):
  """Default getter"""
  try:
    return l[i]
  except IndexError:
    return default_val
  except KeyError:
    return default_val


def run(args):
  log = logging.getLogger(__name__)
  colours = args['--colours']
  axs = []

  for fname in args['<files>']:
    log.info("Processing:" + fname)
    series = {}  # {(ax_num, 'series_name', 'fig_name')
    #            #  : [[x0,y0,z0], ... ], ... }
    series_name = ''
    fig_name = ''
    with (sys.stdin if fname is '-' else open(fname)) as fp:
      for raw_row in fp:
        row = raw_row.strip()
        if not len(row):
          continue
        elif row[0] == '#':  # Comment
          if len(row) > 1:
            if row[1] == '@':  # Series
              series_name = row[2:]
              log.info("Found:Series:" + series_name)
              series[(len(axs), series_name, fig_name)] = []
            elif row[1] == '%':  # Figure
              fig = plt.figure()
              projection = get_options(row).get("projection", "3d").lower()
              projection = "3d" if projection == "3d" else None
              axs.append(fig.add_subplot(111, projection=projection))
              log.info("Found:Figure:" + row[2:])
              fig_name = strip_options(row[2:])
              plt.title(fig_name)
        else:
          series[(len(axs), series_name, fig_name)].append(map(
              lambda x: float(x) if x else 0.0, row.split(',')))

    while len(series) > len(colours):
      colours = colours * 2

    ofname = fn_prepend(fname, 'trans_' +
                        ('R' if args['--right'] else 'L') + '_')
    if args['--trans-mod']:
      log.warn("Applying custom transformations (%s)"
               % ("R" if args['--right'] else "L"))
      my_trans = import_module(args['--trans-mod'])
      my_trans.my_trans(series, isRight=args['--right'])
      save_csv(series, ofname)
    if args['--save-mod']:
      my_save = import_module(args['--save-mod'])
      my_save.my_save(series, ofname.replace('.csv', '.json'),
                      isRight=args['--right'])

    if args['--timeseries']:
      # private axes for components

      log.debug(str([u for v in series.values() for u in v if len(u) > 5]))

      _, _axs = plt.subplots(nrows=max(len(u)
                                       for v in series.values()
                                       for u in v) - 1,
                             ncols=1,
                             sharex=True, squeeze=True)
      plt.xlabel('time')

    for (c, series_name) in zip(colours, series.keys()):
      def p(idx):
        return coord(series[series_name], idx)
      log.info("Processing:" + series_name[1])

      ax = axs[series_name[0] - 1]
      is3d = hasattr(ax, "set_zlabel")

      ax_opts = get_options(series_name[1])

      plot_type = ax_opts.get('plot_type', 'scatter')
      # label = ax_opts.get('label', strip_options(series_name[1]))
      plot_fn = getattr(ax, plot_type)
      plot_kwargs = dict(ax_opts)
      for internalOpt in ("plot_type",):
        plot_kwargs.pop(internalOpt, None)
      xyzuvwlabs = [plot_kwargs.pop(i + "label", i) for i in "xyzuvw"]

      # defaults
      plot_kwargs.setdefault("label", strip_options(series_name[1]))
      plot_kwargs.setdefault("c", c)
      plot_kwargs.setdefault("alpha", 1)

      # special positive defaults
      if plot_type == "scatter":
        plot_kwargs.setdefault('s', 32)
        plot_kwargs['s'] = float(plot_kwargs['s'])
      elif plot_type == "plot":
        plot_kwargs.setdefault("lw", 2)
        plot_kwargs["lw"] = float(plot_kwargs["lw"])
        for unsupportedOpt in ("depthshade",):
          if plot_kwargs.pop(unsupportedOpt, None) is not None:
            log.warn("- %s ignored" % unsupportedOpt)
      elif not is3d:
        for unsupportedOpt in ("depthshade",):
          if plot_kwargs.pop(unsupportedOpt, None) is not None:
            log.warn("- %s ignored" % unsupportedOpt)

      # special negative defaults
      if plot_type not in ("plot",):
        if is3d:
          plot_kwargs.setdefault("depthshade", args['--depthshade'])
          plot_kwargs["depthshade"] = bool(plot_kwargs["depthshade"])

      if args['--timeseries']:
        # TODO: isErr = bool([u for u in series[series_name] if len(u) >= 8])
        isErr = bool([u for u in series[series_name] if len(u) >= 8])
        time_idx = 7 if isErr else min(len(v) for v in series[series_name]) - 1
        coord_indxs = [i for i in
                       range(max(len(v) for v in series[series_name]))
                       if i != time_idx]
        plot_fn(p(get(coord_indxs, 0, -1)),
                p(get(coord_indxs, 1, -1)),
                p(get(coord_indxs, 2, -1)), **plot_kwargs)

        plt.sca(_axs[0])
        plt.title(series_name[2])
        for ylab, indx, ax in zip(xyzuvwlabs, coord_indxs, _axs):
          plt.sca(ax)
          plot_fn = getattr(ax, plot_type)
          plot_fn(p(time_idx), p(indx), **plot_kwargs)
          plt.ylabel(ylab)
          plt.legend()
      else:
        isErr = bool([u for u in series[series_name] if len(u) >= 6])
        if isErr:
          log.info("+ errorbars")
          isPlt = plot_type == "plot"
          plot_kwargs.setdefault("linestyle", ":" if isPlt else "")
          plot_kwargs.setdefault("marker", "" if isPlt else "o")
          # for unsupportedOpt in ("s", "depthshade"):
          #   if plot_kwargs.pop(unsupportedOpt, None) is not None:
          #     log.warn("- %s ignored" % unsupportedOpt)
          errorbar3d(ax, *map(p, range(3)),
                     xerr=p(3), yerr=p(4), zerr=p(5), **plot_kwargs)
        else:
          log.debug(series_name[1])
          plot_fn(*map(p, range(3 if is3d else 2)), **plot_kwargs)

        for (crd, lab) in zip("xyz" if is3d else "xy", xyzuvwlabs):
          log.debug("labels: " + str(xyzuvwlabs))
          getattr(ax, "set_" + crd + "label")(lab)

    for ax in axs:
      plt.sca(ax)
      if args['--fix-aspect']:
        axisEqual3D(ax)
      plt.legend()
      if args['--output-path']:
        ofname = args['--output-path'].rstrip('\\/') + \
            '/' + plt.getp(ax, 'title') + '.png'
        plt.savefig(ofname)
        log.info("Saved to " + ofname)

  plt.show()


def main(argv=None):
  """argv  : list, optional (default: sys.argv[1:])"""
  args = docopt("plot3d v" + __version__ + '\n' +
                __doc__ + __author__ + '\n',
                version=__version__, argv=argv)
  logging.basicConfig(level=getattr(logging, args["--log"], logging.INFO))
  log = logging.getLogger(__name__)
  log.debug(args)
  return run(args) or 0


if __name__ == '__main__':
  main()



import mpl_toolkits.mplot3d.axes3d as axes3d  # NOQA
import logging
__all__ = ["errorbar3d"]
__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
__version__ = "0.1.2"


def errorbar3d(ax, x, y, z, zerr=None, yerr=None, xerr=None, **kwargs):
  """3D plots with errorbars. Inspired by
  http://mple.m-artwork.eu/home/posts/simple3dplotwith3derrorbars
  """
  log = logging.getLogger(__name__)
  plotargs = dict(kwargs)
  plotargs.setdefault("linestyle", "None")
  plotargs.setdefault("marker", 'o')
  barcolour = plotargs.pop("barcolour", 'k')
  log.debug("plotargs:" + str(plotargs))
  if any(scatteropt in plotargs for scatteropt in ("s", "depthshade")):
    log.debug("scatter")
    if plotargs["linestyle"].lower() in ("none", "", " "):
      plotargs.pop("linestyle")
    ax.scatter(x, y, zs=z, **plotargs)
  else:
    log.debug("plot")
    ax.plot(x, y, zs=z, **plotargs)
  # errorbars
  for i in range(len(x)):
    if xerr is not None and xerr[i]:
      ax.plot([x[i] + xerr[i], x[i] - xerr[i]],
              [y[i], y[i]],
              [z[i], z[i]],
              marker="_", c=barcolour)
    if yerr is not None and yerr[i]:
      ax.plot([x[i], x[i]],
              [y[i] + yerr[i], y[i] - yerr[i]],
              [z[i], z[i]],
              marker="_", c=barcolour)
    if zerr is not None and zerr[i]:
      ax.plot([x[i], x[i]],
              [y[i], y[i]],
              [z[i] + zerr[i], z[i] - zerr[i]],
              marker="_", c=barcolour)



from __future__ import division
import pycuda.driver as cuda
import pycuda.autoinit  # NOQA
from pycuda.compiler import SourceModule
import pycuda.gpuarray as gpuarray
import numpy as np
import logging

__all__ = ["cuda", "gpuarray", "CudaFunc3D", "CudaFunc"]
__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
__version__ = "1.0.0"


class CudaFunc3D(object):
  In = cuda.In
  Out = cuda.Out

  @classmethod
  def ctypestr(cls, dtype):
    """Convert a Python data type to a ctype str"""
    res = np.ctypeslib.as_ctypes_type(dtype).__name__
    assert res[:2] == 'c_'
    return res[2:]

  @classmethod
  def ensureContiguousArray(cls, x):
    if not x.flags['C_CONTIGUOUS']:
        log = logging.getLogger(__name__)
        log.warn("making contiguous")
        x = np.ascontiguousarray(x)
    return x

  @classmethod
  def InC(cls, x):
    return x if isinstance(x, cls.In) else \
        cls.In(cls.ensureContiguousArray(x))

  @classmethod
  def OutC(cls, x):
    if isinstance(x, cls.Out):
      return x
    if not x.flags['C_CONTIGUOUS']:
      raise TypeError("Output is not contiguous")
    return cls.Out(x)

  @property
  def size(self):
    return self.size_t(reduce(lambda x, y: x * y, self.shape))

  def __init__(self, shape, block_dim=None, grid_dim=None,
               size_t=np.uint32, wrapper=None):
    r"""PyCUDA function wrapper.
    shape  : 3-tuple (x, y, z) of dimensions (used in grid dim calculation)
    block_dim  : 3-tuple (x, y, z) (default: (1024, 1, 1))
    grid_dim  : 3-tuple (x, y, z)
      (default: (shape + block_dim - 1) / block_dim)
    wrapper  : 2-tuple string added around `build` code
      (default:
        (
          'extern "C" { __global__ void func',
          '}\n'
        )
      )
    """
    self.start = cuda.Event()
    self.end = cuda.Event()
    self.size_t = size_t
    self.shape = size_t(shape)

    if block_dim is None:
      self.block_dim = [1024] + [1] * (len(shape) - 1)
    else:
      self.block_dim = block_dim

    if grid_dim is None:
      self.grid_dim = [(n + b - 1) // b
                       for (n, b) in zip(shape, self.block_dim)]
    else:
      self.grid_dim = grid_dim

    if wrapper is None:
      self.wrapper = ['extern "C" { __global__ void func', "}\n"]
    else:
      self.wrapper = wrapper

  def build(self, code, no_extern_c=1, func='func', wrapper=None, **kwargs):
    """Compile module.
    kwargs  : passed to `SourceModule`.
    """
    log = logging.getLogger(__name__)

    define = dict(
        N_DIM_X=self.shape[0],
        N_DIM_Y=self.shape[1],
        N_DIM_Z=self.shape[2],
        N_DIM=sum(self.shape),
        B_DIM_X=self.block_dim[0],
        B_DIM_Y=self.block_dim[1],
        B_DIM_Z=self.block_dim[2]
    )
    defs = '\n'.join(map(lambda kv: "#define %s %s" % kv, define.items()))

    code = "\n{d}\n{w[0]}{c}{w[1]}".format(
        d=defs, w=wrapper or self.wrapper, c=code)
    log.debug(code)
    self.func = SourceModule(
        code, no_extern_c=no_extern_c, **kwargs).get_function(func)

  def __call__(self, *a, **k):
    self.start.record()
    for (opt, val) in dict(
        block=tuple(self.block_dim),
        grid=tuple(self.grid_dim)
    ).items():
      k.setdefault(opt, val)
    res = self.func(*a, **k)
    self.end.record()
    return res

  def sync(self):
    return self.end.synchronize()

  def elapsed(self):
    return self.start.time_till(self.end) * 1e-3


class CudaFunc(CudaFunc3D):
  def __init__(self, size, block_dim=1024, **k):
    super(CudaFunc, self).__init__(
        (size, 1, 1),
        block_dim=(block_dim, 1, 1),
        grid_dim=((size + block_dim - 1) // block_dim, 1, 1),
        **k)



from multiprocessing import Process, cpu_count
__all__ = ["UnorderedQMap"]
__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
__version__ = "0.1.0"


SENTINEL = "StOp The W0rld"


def worker(inQ, outQ):
  for func, args, kwargs in iter(inQ.get, SENTINEL):
    outQ.put(func(*args, **kwargs))


class UnorderedQMap(object):
  """Maps input Queue of 3-tuples: (function, (*args), {**kwargs})
  onto Queue of results with Context Manager syntax
  Example Usage:

  >>> from multiprocessing import Queue
  >>> def stringify(*a, **k):
  ...   return str((a, k))
  ...
  >>> inQ, outQ = Queue(), Queue()
  >>> for i in range(10):
  ...   inQ.put((stringify, (i, "say"), dict(hello="world"[i % 5])))
  ...
  >>> with UnorderedQMap(outQ, inQ):
  ...   for i in range(10):
  ...     print(outQ.get())
  """
  try:
    CPU_COUNT = cpu_count()
  except Exception:  # pragma: no cover
    CPU_COUNT = 4

  def __init__(self, resultQ, inputQ, processes=None):
    """Careful to put results Q first.
    `max_processes` default: cpu_count()
    """
    self.iQ = inputQ
    self.oQ = resultQ
    self.n = self.CPU_COUNT if processes is None else processes

  def __enter__(self):
    # Start worker processes
    for i in range(self.n):
      Process(target=worker, args=(self.iQ, self.oQ)).start()

  def __exit__(self, *exc):
    for i in range(self.n):
      self.iQ.put(SENTINEL)
    return False



#!/usr/bin/env python
"""showslice v%s
Usage:
  showslice [options] <binfile>

Arguments:
  <binfile>  : data file

Options:
  -v, --version  : print version number and exit.
  -h, --help     : print this help mesasge and exit.
  --log=<lvl>  : CRITICAL|ERROR|WARN(ING)|[default: INFO]|DEBUG|NOTSET
  --parfile=<p>    : (default: <binfile>.rstrip('_1.bin') + '.par')
  --cmap=<c>   : [default: jet]
  --vmax=<v>   : [default: auto]|each_auto|none|<float>
  --vmin=<v>   : [default: auto]|each_auto|none|<float>
  -n=<N>       : int, slice number (default: half way)
  --medfilt=<m>    : median filter size [default: 1:int] is off
  --srmip=<D>  : Short-range MIP half-width depth [default: 0:int]
                 ignored for `--d3`
  --dpi=<D>    : [default: 120:int] for 860 x 540 px figsize
  --out=<dir>  : [default: out]
  --save       : bool, whether to save npz/mat
  --downscale  : bool, whether to downscale by `medfilt` before saving
  --d3  : bool, do 3D projection and use `n` as polar angle in degrees
  --d4  : bool, do both with and without `--d3`
  --anim       : bool, whether to egerate GIF
%s
"""
from __future__ import division

import matplotlib
matplotlib.use("TkAgg")  # NOQA
from matplotlib.backends.backend_tkagg import \
    FigureCanvasTkAgg, NavigationToolbar2TkAgg
# from matplotlib.figure import Figure
import matplotlib.pyplot as plt
import sys
if sys.version_info[0] >= 3:
  import tkinter as tk
  from tkinter import ttk
else:
  import Tkinter as tk
  import Tkinter as ttk

import logging
from argopt import argopt
from tqdm import tqdm, trange
import numpy as np
import re
import os
from ..utils import memoise, TqdmStream, stripEnd

__version__ = "1.2.0"
__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"


RE_PARAMS = re.compile(r"^(\w+)\s+\:\s+(.+)$", flags=re.M)
RE_VALUES = re.compile(
    r'^([0-9]+)\s+(\w+\s+)?"([-\w\s]+)"(\s*right\s*|\s*left\s*)?$', flags=re.M)


def params(parfile):
  log = logging.getLogger(__name__)
  res = dict()
  with open(parfile) as fn:
    raw = fn.read()

    reRes = RE_PARAMS.findall(raw)
    log.debug(reRes)
    res.update([(k.split('(', 1)[0].rstrip('; \t'), v) for (v, k) in reRes])

    reVals = RE_VALUES.findall(raw)
    log.debug(reVals)
    vals = set([int(i[0]) for i in reVals])
    log.debug(vals)
    res["vmax"] = max(vals)
    res["vmin"] = min(vals)

  log.debug(res)
  return res


@memoise()
def knlRotate(D, H, W):
  from caspyr.cuda import CudaFunc3D
  knl = CudaFunc3D([W, H, 1], block_dim=[16, 16, 1])
  knl.build("""
  ( float dst[][N_DIM_Y][N_DIM_X]
  , float src[][N_DIM_Y][N_DIM_X]
  , float ang)
  {
    unsigned x = blockIdx.x * B_DIM_X + threadIdx.x;
    if (x >= N_DIM_X) return;
    unsigned y = blockIdx.y * B_DIM_Y + threadIdx.y;
    if (y >= N_DIM_Y) return;
    unsigned z = blockIdx.z * B_DIM_Z + threadIdx.z;
    if (z >= N_DIM_Z) return;

    float xMid = x - N_DIM_X / 2.0f;
    float yMid = y - N_DIM_Y / 2.0f;
    float s = sin(ang);
    float c = cos(ang);

    int xx = round(N_DIM_X / 2.0f + xMid * c + yMid * s);
    if (0 > xx || xx >= N_DIM_X) return;
    int yy = round(N_DIM_Y / 2.0f - xMid * s + yMid * c);
    if (0 > yy || yy >= N_DIM_Y) return;

    for (unsigned zz = 0; zz < N_REAL_DIM_Z; ++zz)
      dst[zz][y][x] = src[zz][yy][xx];
  }""".replace("N_REAL_DIM_Z", str(D)))
  return knl


def mipRot(rotate, iproj=np.max):
  def inner(vol, iAng, axes=(2, 1), reshape=False, order=0, axis=1):
    return iproj(rotate(
        vol, iAng, axes=axes, reshape=reshape, order=order), axis=axis)
  return inner


@memoise()
def knlRotations(NAngles, D, H, W):
  from caspyr.cuda import CudaFunc3D
  knl = CudaFunc3D([W, H, NAngles], block_dim=[8, 8, 4])
  knl.build("""
  ( float dst[][N_REAL_DIM_Z][N_DIM_Y][N_DIM_X]
  , float src[][N_DIM_Y][N_DIM_X])
  {
    unsigned x = blockIdx.x * B_DIM_X + threadIdx.x;
    if (x >= N_DIM_X) return;
    unsigned y = blockIdx.y * B_DIM_Y + threadIdx.y;
    if (y >= N_DIM_Y) return;
    // z-dimension is actually rotation angle
    unsigned a = blockIdx.z * B_DIM_Z + threadIdx.z;
    if (a >= N_DIM_Z) return;
    float ang = float(a * 2 * 3.1415926535897932384626433832795 / N_DIM_Z);

    float xMid = x - N_DIM_X / 2.0f;
    float yMid = y - N_DIM_Y / 2.0f;
    float s = sin(ang);
    float c = cos(ang);

    int xx = round(N_DIM_X / 2.0f + xMid * c + yMid * s);
    if (0 > xx || xx >= N_DIM_X) return;
    int yy = round(N_DIM_Y / 2.0f - xMid * s + yMid * c);
    if (0 > yy || yy >= N_DIM_Y) return;

    for (unsigned z = 0; z < N_REAL_DIM_Z; ++z)
      dst[a][z][y][x] = src[z][yy][xx];
  }""".replace("N_REAL_DIM_Z", str(D)))
  return knl


def mipRots(rotates, iproj=np.max):
  def inner(vol, nAng, axes=(2, 1), reshape=False, order=0, axis=1):
    return [iproj(v, axis=axis) for v in rotates(
        vol, nAng, axes=axes, reshape=reshape, order=order)]
  return inner


@memoise()
def knlRotationMips(NAngles, D, H, W):
  """
  rotates along Z (D) NAngles times, MIPing each along Y (H)
  """
  from caspyr.cuda import CudaFunc3D
  knl = CudaFunc3D([W, H, NAngles], block_dim=[16, 1, 16])
  knl.build("""
  ( float dst[][N_REAL_DIM_Z][N_DIM_X]
  , float src[N_REAL_DIM_Z][N_DIM_Y][N_DIM_X])
  {
    unsigned x = blockIdx.x * B_DIM_X + threadIdx.x;
    if (x >= N_DIM_X) return;
    unsigned y = blockIdx.y * B_DIM_Y + threadIdx.y;
    if (y >= N_DIM_Y) return;
    // z-dimension is actually rotation angle
    unsigned a = blockIdx.z * B_DIM_Z + threadIdx.z;
    if (a >= N_DIM_Z) return;
    float ang = float(a * 2 * 3.1415926535897932384626433832795 / N_DIM_Z);

    float xMid = x - N_DIM_X / 2.0f;
    float yMid = y - N_DIM_Y / 2.0f;
    float s = sin(ang);
    float c = cos(ang);

    int xx = round(N_DIM_X / 2.0f + xMid * c + yMid * s);
    if (0 > xx || xx >= N_DIM_X) return;
    int yy = round(N_DIM_Y / 2.0f - xMid * s + yMid * c);
    if (0 > yy || yy >= N_DIM_Y) return;

    for (unsigned z = 0; z < N_REAL_DIM_Z; ++z)
      dst[a][z][x] = max(dst[a][z][x], src[z][yy][xx]);
  }""".replace("N_REAL_DIM_Z", str(D)))
  return knl


def rotProject(vols, stepDeg=6, warn_cpu=True, force_cpu=False,
               blocking=False):
  """
  vols   : array of 3D volumes, shape: N, Z, Y, X
  @return out  : array of array of 2D projections, shape:
      N, 360 / stepDeg, Z, X
  """
  from caspyr.multiprocessing import promise_mmap, cpu_count
  from scipy.ndimage import rotate as cpu_rotate
  log = logging.getLogger(__name__)

  nthreads = min(max(1, int(cpu_count() / 2)), len(vols))

  try:
    if force_cpu:
      raise ValueError("cpu")

    warn = [warn_cpu]

    def rotates(arr, NAng, **k):
      if mypy.ndim(arr) != 3:
        raise TypeError("expected 3Dinput")
      if k.get("axes") not in [(1, 2), (2, 1)] or k.get("reshape"):
        if warn[0]:
          raise NotImplementedError("can only roate about z without reshaping")
      im = arr.astype(np.float32).copy('C')
      log.debug("knl I/O shape:%s" % str(im.shape))
      dst = np.zeros((NAng, im.shape[0], im.shape[2]),
                     dtype=im.dtype, order='C')
      # D, H, W = arr.shape
      gRotate = knlRotationMips(NAng, *arr.shape)
      gRotate(gRotate.Out(dst), gRotate.In(im))
      # gRotate.sync()
      return dst

    # foreground GPU
    steps = int(360 / stepDeg)
    log.info("steps:%d" % steps)
    res = [rotates(vol, steps,
                   axes=(2, 1), reshape=False, order=0, axis=1)
           for vol in tqdm(vols, desc="volume", unit="rot", unit_scale=steps)]
  except Exception as e:
    if not (force_cpu and "cpu" == str(e)):
      log.warn("using CPU")
      log.warn(str(e))
    else:
      log.info("using CPU")
    res = None
  else:
    log.info("using GPU")

  if res is not None:
    return res

  # background CPU thread
  res = [
      promise_mmap(mipRot(cpu_rotate),
                   [(vol, iAng) for iAng in np.arange(0, 360, stepDeg)],
                   nthreads=nthreads)
      for vol in vols]

  if blocking:
    # main thread calculates from other end
    [None for vol in tqdm(res, desc="volume")
     for _ in tqdm(reversed(vol), desc="rotating", unit_scale=stepDeg)]
  return res


def getattrN(obj, k, d):
  res = getattr(obj, k, None)
  return d if res is None else res


class StartPage(tk.Frame):
  def __init__(self, parent, controller):
    # log = logging.getLogger(__name__)
    tk.Frame.__init__(self, parent)

    self.args = controller.args
    self.par = controller.par

    self.f = plt.figure(1337, figsize=(8, 4.5),
                        dpi=self.args.dpi, facecolor="black")
    self.canvas = FigureCanvasTkAgg(self.f, self)
    self.canvas.draw()
    self.grid_rowconfigure(0, weight=60)
    self.grid_columnconfigure(0, weight=1)
    self.canvas.get_tk_widget().grid(row=0, column=0, sticky="NSEW")
    # self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True)

    container = tk.Frame(self)
    self.grid_rowconfigure(1, weight=1)
    container.grid(row=1, column=0, sticky="EW")

    if not self.args.anim:
      toolbar = NavigationToolbar2TkAgg(self.canvas, container)
      toolbar.update()
      # self.canvas._tkcanvas.pack(side=tk.TOP, fill=tk.BOTH, expand=True)

    self.bQuit = tk.Button(container, bitmap="error", command=controller.quit)
    # container.grid_columnconfigure(0, weight=1)
    # self.bQuit.grid(row=1, column=1)
    self.bQuit.pack()

    self.container = container
    self.setDat(controller.dat)

  def setDat(self, dat):
    log = logging.getLogger(__name__)
    self.dat = dat
    args = self.args

    plotWidth = plotHeight = 1
    while plotHeight * plotWidth < len(self.dat):
      if plotWidth == plotHeight:
        plotWidth += 1
      else:
        plotHeight += 1
    self.f, axs = plt.subplots(plotHeight, plotWidth,
                               num=1337,
                               sharex=not args.d4, sharey=not args.d4)
    for (k, v) in getattrN(args, "fig_set", {}).items():
      if k == "title":
        self.f.suptitle(v, y=1, verticalalignment="top", color="white")
      else:
        getattr(self.f, k)(v)
    try:
      self.axs = list(axs.flat)
      [(ax.clear(), ax.axis("off")) for ax in self.axs[len(self.dat):]]
      self.axs = list(axs.flat)[:len(self.dat)]
      log.info("%r axes created" % len(self.axs))
    except AttributeError:
      self.axs = [axs]

    self.len = max([len(d) for d in self.dat])
    z = self.len
    log.debug(z)
    if self.args.anim:
      z = 0
    else:
      self.scSlice = ttk.Scale(self.container, from_=0, to=z - 1,
                               resolution=1,
                               orient=ttk.HORIZONTAL,
                               sliderlength=5, length=600,
                               label="angle" if self.args.d3 else "slice",
                               command=self.showslice)
      # container.grid_rowconfigure(2, weight=1)
      # container.grid_columnconfigure(1, weight=7)
      # self.scSlice.grid(row=2, column=0)
      self.scSlice.pack(side=tk.BOTTOM, fill=tk.BOTH, expand=True)
    initSlice = self.args.n if self.args.n >= 0 else z // 2
    log.debug(initSlice)
    if not self.args.anim:
      self.scSlice.set(initSlice)

  def showslice(self, i, draw=True):
    log = logging.getLogger(__name__)
    i = int(i)
    log.debug(i)
    axs = self.axs
    args = self.args
    log.debug("vmin:%r, vmax:%r" % (args.vmin, args.vmax))

    ax_kwargs = getattrN(args, "ax_set", [{} for _ in axs])
    im_kwargs = getattrN(args, "im_set", [{} for _ in axs])
    assert len(self.dat) == len(axs)
    for (j, (vol, ax)) in enumerate(zip(self.dat, axs)):
      # log.debug("ax:%r" % j)
      for k, v in dict(cmap=getattrN(args, "cmap", ["Greys_r"] * (j + 1))[j],
                       aspect="equal", origin="lower",
                       vmin=args.vmin[j],
                       vmax=args.vmax[j]).items():
        im_kwargs[j].setdefault(k, v)
      ax.clear()
      hwidth = getattrN(args, "srmip", 0)  # Short Range MIP half-width
      d4mip = args.d4 and j >= len(axs) // 2
      if hwidth > 0 and not d4mip:
        if args.d3:
          log.warn("expected:srmip==0, got:%s" % hwidth)
        srmip = np.max(vol[max(0, min(len(vol) - 1, i - hwidth)):i + hwidth],
                       axis=0)
      elif hwidth and not d4mip:
        raise ValueError("expected:srmip>=0, got:%s" % hwidth)
      else:
        srmip = vol[min(len(vol) - 1, i)]
      ax.imshow(srmip, **im_kwargs[j])
      if d4mip:
        try:
          if args.full_line:
            raise IndexError
          xmin, xmax = (
              srmip[i] > im_kwargs[j]["vmax"] * .4)\
              .nonzero()[0][[0, -1]] / len(srmip[i])
          if xmax - xmin < 0.1:
            raise IndexError
        except IndexError:
          xmin, xmax = 0, 1
        ax.axhline(i, xmin=xmin, xmax=xmax)
      try:
        for k, v in ax_kwargs[j].items():
          if "label" == k:
            getattr(ax, "set_" + k)(v, color="white")
          elif "title" == k:
            ax.text(len(srmip[0]) / 2, 0, v,
                    color="white", va="top", ha="center")
          else:
            getattr(ax, "set_" + k)(v)
      except IndexError:
        if not d4mip:
          raise
      except Exception as e:
        log.warn(str(e))
      ax.axis("off")
    self.f.tight_layout(pad=0, h_pad=0)
    if draw:
      self.canvas.draw()
    else:
      fname = os.path.join(args.basename, "%04d.png" % i)
      log.debug(fname)
      self.f.savefig(fname, facecolor="black")


class mypy(object):
  """numpy-like functions for arbitrary dimesioned nester iterables"""

  @classmethod
  def ndim(cls, i):
    """Recursively considers i[0] until a scalar is found

    see also: np.ndim
    """
    if hasattr(i, "ndim"):
      return i.ndim

    try:
      return 1 + cls.ndim(i[0])
    except IndexError:
      pass
    except TypeError as e:
      if '__getitem__' not in str(e):
        raise
    return 0

  @classmethod
  def vec_reduce(cls, func, i):
    """Recursively considers i[0] until a vector is found

    see also: np.ndim
    """
    ndim = cls.ndim(i)
    if ndim <= 1:
      return func(i)
    else:
      return func([cls.vec_reduce(func, j) for j in i])

  @classmethod
  def min(cls, i):
    return cls.vec_reduce(min, i)

  @classmethod
  def max(cls, i):
    return cls.vec_reduce(max, i)


class MainApp(tk.Tk):
  def __init__(self, args, dat=None, datStep=None, *a, **k):
    log = logging.getLogger(__name__)
    tk.Tk.__init__(self, *a, **k)
    tk.Tk.wm_title(self, args.parfile)
    bname = "out"
    if args.parfile:
      bname = stripEnd(args.parfile, ".par")
    elif args.binfile:
      bname = stripEnd(args.binfile, "_1.bin")
    args.basename = os.path.join(
        os.path.dirname(bname),
        args.out if args.out is not None else "out",
        os.path.basename(bname))
    self.args = args
    if datStep is not None:
      raise log.warn("DeprecationWarning:use 4D input instead of datStep")
    if dat is not None:
      dat_ndim = mypy.ndim(dat)
      log.debug("dat:ndim:%d" % dat_ndim)
      if dat_ndim == 2:
        dat = dat.reshape((1, 1, dat.shape[0], dat.shape[1]))
      elif dat_ndim == 3:
        dat = [dat[i::datStep] for i in range(datStep)]
      elif dat_ndim == 4:
        if datStep is not None and datStep != 1:
          raise ValueError(
              "expected:datStep=1, got:%r since dat.ndim==4", datStep)
      else:
        raise ValueError(
            "expected:`dat.ndim in [2, 3, 4]`, got:%d" % dat.ndim)
      self.dat = dat
      z = len(dat[0])
      w = len(dat[0][0])
      # assert w == dat.shape[2]
      assert (datStep is None) or len(dat) == datStep
      self.par = {"array_size": w, "end_slice": z, "start_slice": 0,
                  "vmin": mypy.min(dat) if args.vmin == "auto" else -1,
                  "vmax": mypy.max(dat) if args.vmin == "auto" else -1}
      log.debug("global vmin/max")
    else:
      self.par = params(args.parfile)
      w = int(self.par["array size"])
      z = int(self.par["end_slice"]) - int(self.par["start_slice"]) + 1
      self.dat = np.fromfile(args.binfile, dtype=np.float32)
      log.info(len(self.dat))
      self.dat.resize((1, z, w, w))
      log.info(args.parfile + ':' + 'x'.join(map(str, self.dat.shape)))

    if args.medfilt > 1:
      from scipy.ndimage.filters import median_filter
      self.dat = median_filter(self.dat, size=tuple([1] + [args.medfilt] * 3))

    if "each" in getattrN(args.vmin, "lower", lambda: "")():
      self.args.vmin = [np.min(d) for d in self.dat]
    self.argsAxisListCast("vmin")

    if "each" in getattrN(args.vmax, "lower", lambda: "")():
      self.args.vmax = [np.max(d) for d in self.dat]
    self.argsAxisListCast("vmax")

    if ',' in args.cmap:
      args.cmap = args.cmap.split(',')
    self.argsAxisListCast("cmap", dtype=str)

    # 2D slices -> 360 degrees 2D projections of 3D volumes
    if args.d3:
      self.dat = rotProject(self.dat)
    elif args.d4:
      self.dat.extend(rotProject(self.dat, stepDeg=360 / len(self.dat[0])))
      self.args.vmax.extend(self.args.vmax)
      self.args.vmin.extend(self.args.vmin)
      self.args.cmap.extend(self.args.cmap)

    if args.save and not args.anim:
      sdat = self.dat
      if args.downscale and args.medfilt > 1:
        from skimage.transform import downscale_local_mean
        sdat = downscale_local_mean(sdat, tuple([1] + [args.medfilt] * 3))
      log.info("save:" + args.basename + ".npz")
      np.savez_compressed(args.basename, sdat)
      log.info("save:" + args.basename + ".mat")
      # from scipy.io import savemat
      # savemat(bname, {"ph": sdat})
      from hdf5storage import write as savemat
      savemat({u"ph": np.asarray(sdat)}, '.', args.basename + ".mat",
              matlab_compatible=True, truncate_existing=True)

    container = tk.Frame(self)
    container.pack(side="top", fill="both", expand=True)
    container.grid_rowconfigure(0, weight=1)
    container.grid_columnconfigure(0, weight=1)

    self.frames = {}
    self.frames[StartPage] = StartPage(container, self)
    self.frames[StartPage].grid(row=0, column=0, sticky="nsew")
    f = self.show_frame(StartPage)
    if args.anim:
      if args.save:
        try:
          os.makedirs(args.basename)
        except:
          pass
      # silly fix moving axes
      for _ in range(10):
        f.showslice(0, draw=not args.save)
      for i in trange(f.len, unit="frame", disable=not args.save):
        f.showslice(i, draw=not args.save)
      if args.save:
        cmd = "ffmpeg -i {0}/%04d.png {0}.mp4".format(args.basename)
        log.info(cmd)
        os.system(cmd)
      f.showslice(f.len // 2)

  def argsAxisListCast(self, par, dtype=float):
    """
    par  : self.par key for "auto" (default)
    """
    log = logging.getLogger(__name__)
    val = getattr(self.args, par)
    valStr = getattrN(val, "lower", lambda: "")()
    n = len(self.dat)
    if val is None or valStr == "none":
      setattr(self.args, par, [None] * n)
    elif valStr == "auto":
      setattr(self.args, par, [self.par[par]] * n)
    elif hasattr(val, "__iter__"):
      pass  # setattr(self.args, par, val)
    else:
      setattr(self.args, par, [dtype(val)] * n)
    log.info("%s:%s" % (par, getattr(self.args, par)))

  def show_frame(self, cont):
    log = logging.getLogger(__name__)
    log.debug(cont.__name__)
    frame = self.frames[cont]
    frame.tkraise()
    return frame


def run(args):
  """Non-TK version"""
  log = logging.getLogger(__name__)
  dat = np.fromfile(args.binfile, dtype=np.float32)
  par = params(args.parfile)
  w = int(par["array size"])
  z = int(par["end_slice"]) - int(par["start_slice"]) + 1
  log.info("%s:%dx%dx%d" % (args.parfile, w, w, z))
  dat = dat.reshape((z, w, w))[args.n if args.n >= 0 else z // 2]
  if args.medfilt > 1:
    from scipy.ndimage.filters import median_filter
    dat = median_filter(dat, size=args.medfilt)
  log.info("slice:" + 'x'.join(map(str, dat.shape)))
  plt.imshow(dat, cmap="jet", aspect="equal", origin="lower")
  plt.axis("off")
  plt.show()
  if args.save:
    sdat = dat
    if args.downscale and args.medfilt > 1:
      from skimage.transform import downscale_local_mean
      sdat = downscale_local_mean(sdat, tuple([args.medfilt] * 3))
    np.save(stripEnd(args.parfile, ".par"), sdat)


def parseArgv(argv, configLog=True):
  args = argopt(__doc__ % (__version__, __author__),
                version=__version__).parse_args(args=argv)
  # args = DictAttrWrap(docopt(__doc__ % (__version__, __author__),
  #                            version=__version__, argv=argv))
  logging.basicConfig(level=getattr(logging, args.log, logging.INFO),
                      stream=TqdmStream)
  log = logging.getLogger(__name__)
  if not args.parfile:
    args.parfile = stripEnd(args.binfile, '_1.bin') + '.par'
    log.info("parfile:" + args.parfile)

  args.n = int(args.n) if args.n is not None else -1
  log.debug(args)
  return args


def main(argv=None):
  """argv  : list, optional (default: sys.argv[1:])"""
  args = parseArgv(argv)
  # return run(args) or 0
  app = MainApp(args)
  app.mainloop()


if __name__ == "__main__":
  main()



#!/usr/bin/env python
"""pusher v%s
Usage:
  pusher [options] <title>

Options:
  -b BODY, --body BODY  Message text (default: {id}|{time})
  -i ID, --id ID        Message topic or device ID [default: news]
  -e SEC, --expiry SEC  Seconds before message expires [default: 60*60*24:int]
  --log LEVEL           CRITICAL|ERROR|WARN(ING)|[default: INFO]|DEBUG|NOTSET
%s
"""
from argopt import argopt
from os import getenv
import requests
from time import ctime
import logging

__all__ = ["run", "RunArgs", "script", "main"]
__version__ = "0.3.1"
__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
URL = "https://fcm.googleapis.com/fcm/send"


def run(args):
  """@param args: RunArgs"""
  log = logging.getLogger(__name__)

  if not args.id.startswith('/') and len(args.id) < 99:
    # inject short topic id into message body if necessary
    if args.body is None:
      args.body = str(args.id)
    else:
      args.body = args.body.replace("{id}", args.id)
    # prefix /topics/
    args.id = "/topics/" + args.id

  # long id with no body -> time
  if args.body is None:
    args.body = "{time}"
  args.body = args.body.\
      replace("{time}", ctime()).\
      replace("{title}", args.title)
  log.debug(args)

  auth_key = getenv("FCM_PUSHER_AUTH_KEY", None)
  if auth_key is None:
    raise EnvironmentError("Please set FCM_PUSHER_AUTH_KEY and try again")

  headers = dict(Authorization="key=" + auth_key)
  data = dict(to=args.id, priority="high", time_to_live=args.expiry,
              notification=dict(title=args.title,
                                body=args.body,
                                sound="default"))
  log.debug(str(headers))
  log.debug(str(data))
  r = requests.post(URL, headers=headers, json=data)
  log.info(r.content)
  return r.json()


class RunArgs(object):
  def __init__(self, title, body=None, id="news", expiry=60 * 60 * 24):
    """Default arguments for run()"""
    self.body = body
    self.id = id
    self.expiry = expiry
    self.title = title


def script(*args, **kwargs):
  """alias for run(RunArgs(...))"""
  return run(RunArgs(*args, **kwargs))


def main(argv=None):
  """argv  : list, optional (default: sys.argv[1:])"""
  args = argopt(__doc__ % (__version__, __author__),
                version=__version__).parse_args(args=argv)
  logging.basicConfig(level=getattr(logging, args.log, logging.INFO))
  log = logging.getLogger(__name__)
  log.debug(args)
  return run(args) or 0


if __name__ == "__main__":
  main()



#!/usr/bin/env python
"""duplicate v%s

A python file (which generates itself) to get you started on new scripts.

Usage:
  duplicate [options] [<output>]

Options:
  -f, --force  : Force overwriting files.
  --log LEVEL  : (FAT|CRITIC)AL|ERROR|WARN(ING)|[default: INFO]|DEBUG|NOTSET

Arguments:
  [<output>]  : Output duplicate file name (default: self~).
                Use '-' for stdout.

%s
"""
import sys
from io import open as io_open
from os import path
import logging
from argopt import argopt
from ..utils import inputPrompt, stripEnd
__all__ = ["main", "run"]
__version__ = "0.0.0"
__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"


def run(args):
  """@param args: RunArgs"""
  log = logging.getLogger(__name__)

  this_file_raw = __file__
  for ext in ('c', 'o', ''):
    this_file_raw = stripEnd(this_file_raw, ".py" + ext)

  with io_open(this_file_raw + ".py") as fd:
    this_file_raw = fd.read()

  if path.isfile(args.output):
    log.warn("file %s exists." % args.output)
    if not (args.force or inputPrompt("Overwrite")):
      return -1
    log.info("Overwriting " + args.output)

  with (sys.stdout if args.output == '-' else io_open(args.output, 'w')) as fd:
    fd.write(this_file_raw)


def main(argv=None):
  """argv  : list, optional (default: sys.argv[1:])"""
  args = argopt(__doc__ % (__version__, __author__),
                version=__version__).parse_args(args=argv)
  logging.basicConfig(level=getattr(logging, args.log, logging.INFO))
  log = logging.getLogger(__name__)
  if args.output is None:
    args.output = sys.argv[0] + '~'
  log.debug(args)
  return run(args) or 0


if __name__ == "__main__":
  main()



def test_errorbar3d():
  """Test 3D errorbar plotting"""
  from caspyr.plotting import errorbar3d
  import matplotlib.pyplot as plt
  import numpy as np

  fig = plt.figure(dpi=100)
  ax = fig.add_subplot(111, projection='3d')

  # data
  x = sorted(np.random.random(9))
  y = sorted(np.random.random(9))
  z = sorted(np.random.random(9))
  sx = np.random.random(9) / 10
  sy = np.random.random(9) / 10
  sz = np.random.random(9) / 10

  # optional configure axes
  ax.set_xlim3d(0, 1)
  ax.set_ylim3d(0, 1)
  ax.set_zlim3d(0, 1)
  ax.set_xlabel('x')
  ax.set_ylabel('y')
  ax.set_zlabel('z')

  errorbar3d(ax, x, y, z, sz, sy, sx, linestyle=":")
  plt.show()


def test_Plt():
  """Test Plt"""
  from caspyr.plotting import Plt
  import matplotlib.pyplot as pyplot

  Plt.set_dims(2, 2)  # optionally set subplot dimensions
  Plt.fig()
  pyplot.plot(range(10))
  Plt.fig()
  pyplot.plot(range(10, 0, -1))
  Plt.set_dims()  # back to figure mode
  Plt.fig()
  pyplot.plot([1] * 5 + [2] * 5)
  Plt.set_dims(2, 2)  # optionally set subplot dimensions
  Plt.fig()
  pyplot.plot(range(10, 0, -1))
  pyplot.show()



r"""
>>> import sys
>>> from time import time, sleep
>>>
>>> def prnt(*a):
...   pass  # sys.stderr.write(' '.join(map(str, a)))
...
>>> def stringify(*a):
...   sleep(0.1)
...   return str(a)
...
>>> inQ = []
>>> for i in range(10):
...   inQ.append(((i, "say"), dict(hello="world"[i % 5])))
...
>>> tic1 = time()
>>> res1 = promise_map(stringify, inQ)
>>>
>>> # should increase "new" time by 0.2s
>>> prnt(8, res1[8])
>>> prnt(8, res1[8])
>>> prnt(8, res1[8])
>>> prnt(7, res1[7])
>>>
>>> # old version ("new" should finish during this time)
>>> tic2 = time()
>>> res2 = map(stringify, inQ)
>>> tic2 = time() - tic2
>>>
>>> # should not increase "new" time
>>> prnt(9, res1[9])
>>>
>>> sys.stderr.write("old: %.3gs\n" % tic2)
>>> for i in range(len(res1)):
...   prnt(i, res1[i])
...
>>> tic1 = time() - tic1
>>> sys.stderr.write("new: %.3gs\n" % (tic1 - tic2))
"""
from threading import Thread
from time import time
import logging
from ._multiprocessing import cpu_count
from tqdm import tqdm


__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
__version__ = "0.1.0"


class Mapper(object):
  @classmethod
  def map(cls, func, seq, **k):
    """Non-blocking concurrent version of map(func, seq).

    Returns a list-like object which if necessary blocks and
    forces calculation upon individual element access.
    """
    return cls([(func, a, {}) for a in seq], **k)

  def __getitem__(self, i):
    raise NotImplementedError

  def __len__(self):
    raise NotImplementedError

  def __iter__(self):
    for i in range(len(self)):
      yield self[i]
    raise StopIteration
    return

  def __reversed__(self):
    for i in reversed(range(len(self))):
      yield self[i]
    raise StopIteration
    return


class PromiseMap(Thread, Mapper):
  def __init__(self, func_args_kwargs, nostart=False):
    """func_args_kwargs  : 3-tuple, (func, (*args), {**kwargs})"""
    Thread.__init__(self)
    self.daemon = True  # kill thread when main killed (KeyboardInterrupt)
    self.was_killed = False
    self._fak = func_args_kwargs
    self._results = [None] * len(func_args_kwargs)
    self.started = time()
    if not nostart:
      self.start()

  def exit(self):
    self.was_killed = True
    # self.join()  # DO NOT, blocking event, slows down at closing
    return self.report()

  def run(self):
    log = logging.getLogger(__name__)
    for i in range(len(self)):
      # Quit if killed
      # if self.exit_event.is_set():  # TODO: should work but does not...
      if self.was_killed:
        self.was_killed = time() - self.started
        return
      self[i]
    self.was_killed = time() - self.started
    log.info("delivered:" + tqdm.format_interval(self.was_killed))

  def report(self):
    # return self.is_alive()  # TODO: does not work...
    return not self.was_killed

  def __getitem__(self, i):
    if isinstance(i, slice):
      return [self[j] for j in range(len(self))[i]]
    if self._results[i] is None:
      self._results[i] = self._fak[i][0](*self._fak[i][1], **self._fak[i][2])
    return self._results[i]

  def __len__(self):
    return len(self._fak)


promise_map = PromiseMap.map


class PromiseMultiMap(Mapper):
  def __init__(self, func_args_kwargs, nthreads=None, **k):
    """nthreads  : int, (default: caspyr.multiprocessing.cpu_count())"""
    nthreads = nthreads or cpu_count()
    self._threads = [PromiseMap(func_args_kwargs[i::nthreads], **k)
                     for i in range(nthreads)]

  def __getitem__(self, i):
    if isinstance(i, slice):
      return [self[j] for j in range(len(self))[i]]
    j, tid = divmod(i, len(self._threads))
    return self._threads[tid][j]

  def __len__(self):
    return sum(map(len, self._threads))


promise_mmap = PromiseMultiMap.map



"""Usage:
  caspyr (--help|<script>) [options] [<args>...]

Options:
  -h, --help     : Print this help and exit.
  -v, --version  : Print module version number and exit.

Arguments:
  <script>  : %s
"""
from __future__ import print_function
import sys
from importlib import import_module
from glob import glob
from os.path import join, dirname, abspath
__all__ = ["main"]
__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
__version__ = "1.1.0"


def main():
    positionals = filter(lambda a: not a.startswith('-'), sys.argv[1:])
    if not positionals:
        if any(h in sys.argv[1:] for h in ('-h', '--help')):
            pkgroot = abspath(dirname(__file__))
            scripts = [s.replace(join(pkgroot, "scripts", ""), "")[:-3]
                       for s in glob(join(pkgroot, "scripts", "*.py"))]
            scripts = [s for s in scripts if "__" not in (s[:2], s[-2:])]
            print(__doc__ % '|'.join(scripts), end="")
        elif any(v in sys.argv[1:] for v in ('-v', '--version')):
            from ._version import __version__ as modversion
            print(modversion + '-' + __version__)
        else:
            print(__doc__.split("\n\n", 1)[0])
    else:
        script = import_module('.scripts.' + positionals[0], package='caspyr')
        script.main(sys.argv[2:])


if __name__ == "__main__":
    main()



import re
from copy import deepcopy

__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
__version__ = "0.0.0"


class Globber(object):
    def __init__(self, *parts):
        """
        parts  : (tuple(name, regex, glob), ...)
        """
        self.partsKeys = [i[0] for i in parts]
        self.partsRegs = [i[1] for i in parts]
        self.partsGlob = [i[2] for i in parts]
        self.val = self.parseRe()
        self.valRe = re.compile(self.val)
        self.glb = self.parseGlob()
        # glob all files
        self.files = self.glob(self.glb)
        # ensure files pass the regex
        self.files = self.filterRe()

        parts = map(self.split, self.files)
        self.partsVals = [sorted({p[i] for p in parts})
                          for i in range(len(parts[0]))]

    def filterGlob(self, **keys):
        """return self.files that match given key=glob pairs"""
        raise DeprecationWarning
        p = deepcopy(self.partsGlob)
        for k, v in keys.items():
            p[self[k]] = v
        return self.filterRe(self.glob(self.parseGlob(p)))

    def filterRe(self, escape=False, **keys):
        """return self.files that match given key=regex pairs"""
        p = deepcopy(self.partsRegs)
        for k, v in keys.items():
            p[self[k]] = self.escape(v) if escape else v
        r = re.compile(self.parseRe(p))
        return filter(r.match, self.files)

    def filterReUnlike(self, fname, escape=False, **keys):
        """return self.files that are like `fname` but with given key=regex pairs

        Use `None` for regex for default (all)"""
        p = self.split(fname, escape=True)
        for k, v in keys.items():
            i = self[k]
            p[i] = self.partsRegs[i] if v is None \
                else self.escape(v) if escape else v
        r = re.compile(self.parseRe(p))
        return filter(r.match, self.files)

    def filterReLike(self, fname, *keys):
        """return self.files that are like `fname`'s `key`s"""
        p = self.split(fname, escape=True)
        d = {k: p[self[k]] for k in keys}
        return self.filterRe(**d)

    def index(self, key):
        return self.partsKeys.index(key)

    def __getitem__(self, key):
        return self.index(key)

    def parseGlob(self, parts=None):
        return ''.join(self.partsGlob if parts is None else parts)

    def parseRe(self, parts=None):
        return '(' + ')('.join(self.partsRegs if parts is None else parts) + ')'

    def split(self, fname, escape=False):
        """decompose `fname` into its parts"""
        parts = self.valRe.findall(fname)[0]
        return map(self.escape, parts) if escape else list(parts)

    @classmethod
    def glob(cls, *a, **k):
        """sorted version of glob.glob"""
        from glob import glob as g
        return sorted(g(*a, **k))

    @classmethod
    def escape(cls, s):
        """escape filename characters which are valid regex"""
        for c in ".+[]{}()^$":
            s = s.replace(c, '\\' + c)
        return s


def glob(*a, **k):
    return Globber.glob(*a, **k)



import logging
import contextlib

__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
__version__ = "0.0.0"


@contextlib.contextmanager
def getLoggerLevel(name=None, level=None):
    """
    Context yielding `getLogger(name)` with the specified effective `leve1`
    """
    log = logging.getLogger(name)
    old = log.getEffectiveLevel()
    try:
        if level is not None:
            log.setLevel(level)
        yield log
    finally:
        log.setLevel(old)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
from io import open as io_open
try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup

    def find_packages():
        # os.walk -> list[(dirname, list[subdirs], list[files])]
        return [folder.replace("/", ".").lstrip(".")
                for (folder, _, fils) in os.walk(".") if "__init__.py" in fils]

# Main setup.py config #

# Get version from caspyr/_version.py
__version__ = None
version_file = os.path.join(os.path.dirname(__file__), 'caspyr', '_version.py')
with io_open(version_file, mode='r') as fd:
    exec(fd.read())

# Python package config #

extras_require = dict(
    io=['h5py', 'numpy'],
    cuda=['pycuda', 'numpy'],
    plot=['matplotlib'],
    # utils
    prof=['line_profiler'], term=['tqdm'],
    # test suite
    test=['nose', 'flake8'])

README_rst = None
fndoc = os.path.join(os.path.dirname(__file__), 'README.rst')
with io_open(fndoc, mode='r', encoding='utf-8') as fd:
    README_rst = fd.read()

setup(
    name='caspyr',
    version=__version__,
    description="Cinchingly awesome scripts of Pythonic rapture",
    license='MPLv2.0',
    author='Casper da Costa-Luis',
    author_email='casper.dcl@physics.org',
    url='https://github.com/casperdcl/caspyr',
    platforms=['any'],
    packages=find_packages(),
    # install_requires=['docopt', 'argopt'],
    extras_require=extras_require,
    entry_points={'console_scripts': ['caspyr=caspyr._main:main'], },
    long_description=README_rst,
    classifiers=[
        # Trove classifiers
        # (https://pypi.python.org/pypi?%3Aaction=list_classifiers)
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: MacOS X',
        'Environment :: Other Environment',
        'Environment :: Win32 (MS Windows)',
        'Environment :: X11 Applications',
        'Framework :: IPython',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Other Audience',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: BSD',
        'Operating System :: POSIX :: BSD :: FreeBSD',
        'Operating System :: POSIX :: Linux',
        'Operating System :: POSIX :: SunOS/Solaris',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: IronPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Desktop Environment',
        'Topic :: Education :: Computer Aided Instruction (CAI)',
        'Topic :: Education :: Testing',
        'Topic :: Office/Business',
        'Topic :: Other/Nonlisted Topic',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Pre-processors',
        'Topic :: Software Development :: User Interfaces',
        'Topic :: System :: Installation/Setup',
        'Topic :: System :: Logging',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Shells',
        'Topic :: Terminals',
        'Topic :: Utilities'
    ],
    keywords='utilities',
    test_suite='nose.collector',
    tests_require=extras_require['test'] + ['coverage'],
)



from ._memoise import memoise, memo_db, load_db, save_db, hashForce
from ._profile import cprof, lprof
from ._term import sh, TqdmStream, stripEnd, inputPrompt
from ._globber import Globber, glob
from ._table import table
from ._logging import getLoggerLevel

from ..io._h5reader import H5Reader as H
class H5Reader(H):
    def __init__(self, *a,**k):
        raise DeprecationWarning("`H5Reader` moved to `caspyr.io` from `caspyr.utils`")
        super(H5Reader, self).__init__(*a, **k)

__all__ = ["memoise", "memo_db", "load_db", "save_db", "hashForce",
           "cprof", "lprof", "H5Reader", "sh", "TqdmStream", "stripEnd",
           "inputPrompt", "Globber", "glob", "table", "getLoggerLevel"]



from subprocess import Popen, PIPE

__author__ = "Casper da Costa-Luis <casper.dcl@physics.org>"
__version__ = "0.2.0"


def sh(*cmd, **kwargs):
  return Popen(cmd, stdout=PIPE,
               **kwargs).communicate()[0].decode("utf-8")


class TqdmStream(object):
  @classmethod
  def write(cls, msg):
    from tqdm import tqdm
    tqdm.write(msg, end='')


def stripEnd(s, ext):
  """Usage:
  >>> stripEnd("some..ext", ".ext")
  "some."
  """
  return s[:-len(ext)] if s.endswith(ext) else s


def inputPrompt(msg="Continue", mid=' ',
                yes="yY", no="nN", end="? ", default=None):
  """
  @param yes  : str, characters to accept for returning `True`.
  @param no   : str, characters to accept for returning `False`.
  @param end  : str, terminating message.
  @param default  : bool or None, if [default: None] will loop until valid input.

  example format for defaults:
      `{msg}{mid}{yes}/{no}{end}`
  produces:
      `Continue yY/nN? `
  """

  if set(yes).intersection(no):
    raise ValueError("Yes/No options overlap")

  prompt = msg + mid
  prompt += "{}/{}".format(yes, no) if default is None else \
      "[{}]{}/{}".format(yes[0], yes[1:], no) if default else \
      "{}/[{}]{}".format(yes, no[0], no[1:])
  prompt += end

  while True:
    a = raw_input(prompt)
    if a in yes:
      return True
    if a in no:
      return False
    if default is not None:
      return default  # TODO: enforce bool?



from ._multiprocessing import UnorderedQMap
from multiprocessing import cpu_count, freeze_support, Queue, Manager, Lock
from ._promises import PromiseMap, promise_map, PromiseMultiMap, promise_mmap

__all__ = ["UnorderedQMap", "cpu_count",
           "freeze_support", "Queue", "Manager", "Lock",
           "PromiseMap", "promise_map", "PromiseMultiMap", "promise_mmap"]
